#ifndef OPENTK_TIME_H_
#define OPENTK_TIME_H_

#include <opentk/OpenTk_Common.h>

/**
 * This time class is based on a monotonic clock.
 */


struct Time;
typedef struct Time Time;

extern void fhgfs_Time_init(Time* this);
extern void fhgfs_Time_initFromOther(Time* this, Time* other);
extern Time* fhgfs_Time_construct(void);
extern void fhgfs_Time_uninit(Time* this);
extern void fhgfs_Time_destruct(Time* this);

extern void fhgfs_Time_setToNow(Time* this);

static inline void fhgfs_Time_setFromOther(Time* this, Time* other);
static inline int fhgfs_Time_elapsedSinceMS(Time* this, Time* earlierT);
static inline int fhgfs_Time_elapsedSinceMicro(Time* this, Time* earlierT);
static inline int fhgfs_Time_elapsedMS(Time* this);

// operators
static inline fhgfs_bool fhgfs_Time_equals(Time* this, Time* other);
static inline fhgfs_bool fhgfs_Time_notequals(Time* this, Time* other);


struct Time
{
   fhgfs_timespec now;
};


void fhgfs_Time_setFromOther(Time* this, Time* other)
{
   this->now = other->now;
}

//EXPORT_SYMBOL(fhgfs_Time_setFromOther);


int fhgfs_Time_elapsedSinceMS(Time* this, Time* earlierT)
{
   int secs = (this->now.tv_sec - earlierT->now.tv_sec) * 1000;
   int micros = (this->now.tv_nsec - earlierT->now.tv_nsec) / 1000000;

   return secs + micros;
}

//EXPORT_SYMBOL(fhgfs_Time_elapsedSinceMS);


int fhgfs_Time_elapsedSinceMicro(Time* this, Time* earlierT)
{
   int secs = (this->now.tv_sec - earlierT->now.tv_sec) * 1000000;
   int micros = (this->now.tv_nsec - earlierT->now.tv_nsec) / 1000;

   return secs + micros;
}

//EXPORT_SYMBOL(fhgfs_Time_elapsedSinceMicro);


int fhgfs_Time_elapsedMS(Time* this)
{
   int elapsedMS;

   Time currentT;
   fhgfs_Time_init(&currentT);

   elapsedMS = fhgfs_Time_elapsedSinceMS(&currentT, this);

   fhgfs_Time_uninit(&currentT);

   return elapsedMS;
}

//EXPORT_SYMBOL(fhgfs_Time_elapsedMS);


fhgfs_bool fhgfs_Time_equals(Time* this, Time* other)
{
   return (this->now.tv_nsec == other->now.tv_nsec) && (this->now.tv_sec == other->now.tv_sec);
}

//EXPORT_SYMBOL(fhgfs_Time_equals);


fhgfs_bool fhgfs_Time_notequals(Time* this, Time* other)
{
   return (this->now.tv_nsec != other->now.tv_nsec) || (this->now.tv_sec != other->now.tv_sec);
}

//EXPORT_SYMBOL(fhgfs_Time_notequals);


#endif /* OPENTK_TIME_H_ */
