#ifndef SERIALIZATION_H_
#define SERIALIZATION_H_

#include <common/Common.h>

// inliners
static inline unsigned Serialization_serializeInt(char* buf, int value);
static inline fhgfs_bool Serialization_deserializeInt(const char* buf, size_t bufLen,
   int* outValue, unsigned* outLen);
static inline unsigned Serialization_serialLenInt(void);
static inline unsigned Serialization_serializeUInt(char* buf, unsigned value);
static inline fhgfs_bool Serialization_deserializeUInt(const char* buf, size_t bufLen,
   unsigned* outValue, unsigned* outLen);
static inline unsigned Serialization_serialLenUInt(void);
static inline unsigned Serialization_serializeInt64(char* buf, int64_t value);
static inline fhgfs_bool Serialization_deserializeInt64(const char* buf, size_t bufLen,
   int64_t* outValue, unsigned* outLen);
static inline unsigned Serialization_serialLenInt64(void);
static inline unsigned Serialization_serializeUInt64(char* buf, uint64_t value);
static inline fhgfs_bool Serialization_deserializeUInt64(const char* buf, size_t bufLen,
   uint64_t* outValue, unsigned* outLen);
static inline unsigned Serialization_serialLenUInt64(void);




// byte-order transformation

static inline unsigned Serialization_htonlTrans(unsigned value)
{
   return __cpu_to_be32(value);
}

static inline unsigned Serialization_ntohlTrans(unsigned value)
{
   return __be32_to_cpu(value);
}

static inline unsigned short Serialization_htonsTrans(unsigned short value)
{
   return __cpu_to_be16(value);
}

static inline unsigned short Serialization_ntohsTrans(unsigned short value)
{
   return __be16_to_cpu(value);
}

static inline uint64_t Serialization_htonllTrans(uint64_t value)
{
   return __cpu_to_be64(value);
}

static inline uint64_t Serialization_ntohllTrans(uint64_t value)
{
   return __be64_to_cpu(value);
}


// int (de)serialization
static inline unsigned Serialization_serialLenInt(void)
{
   return Serialization_serialLenUInt();
}

static inline unsigned Serialization_serializeInt(char* buf, int value)
{
   return Serialization_serializeUInt(buf, value);
}

static inline fhgfs_bool Serialization_deserializeInt(const char* buf, size_t bufLen,
   int* outValue, unsigned* outLen)
{
   return Serialization_deserializeUInt(buf, bufLen, (unsigned*)outValue, outLen);
}


// uint (de)serialization
static inline unsigned Serialization_serializeUInt(char* buf, unsigned value)
{
   *(unsigned*)buf = __cpu_to_le32(value);

   return Serialization_serialLenUInt();
}

static inline fhgfs_bool Serialization_deserializeUInt(const char* buf, size_t bufLen,
   unsigned* outValue, unsigned* outLen)
{
   unsigned value;

   if(unlikely(bufLen < Serialization_serialLenUInt() ) )
      return fhgfs_false;

   *outLen = Serialization_serialLenUInt();

   value = *(const unsigned*)buf;
   *outValue = __le32_to_cpu(value);

   return fhgfs_true;
}

static inline unsigned Serialization_serialLenUInt(void)
{
   return sizeof(unsigned);
}

// int64 (de)serialization
static inline unsigned Serialization_serializeInt64(char* buf, int64_t value)
{
   return Serialization_serializeUInt64(buf, value);
}

static inline fhgfs_bool Serialization_deserializeInt64(const char* buf, size_t bufLen,
   int64_t* outValue, unsigned* outLen)
{
   return Serialization_deserializeUInt64(buf, bufLen, (uint64_t*)outValue, outLen);
}

static inline unsigned Serialization_serialLenInt64(void)
{
   return Serialization_serialLenUInt64();
}

// uint64 (de)serialization
static inline unsigned Serialization_serializeUInt64(char* buf, uint64_t value)
{
   *(uint64_t*)buf = __cpu_to_le64(value);

   return Serialization_serialLenUInt64();
}

static inline fhgfs_bool Serialization_deserializeUInt64(const char* buf, size_t bufLen,
   uint64_t* outValue, unsigned* outLen)
{
   uint64_t value;

   if(unlikely(bufLen < Serialization_serialLenUInt64() ) )
      return fhgfs_false;

   *outLen = Serialization_serialLenUInt64();

   value = *(const uint64_t*)buf;
   *outValue = __le64_to_cpu(value);

   return fhgfs_true;
}

static inline unsigned Serialization_serialLenUInt64(void)
{
   return sizeof(uint64_t);
}


#endif /*SERIALIZATION_H_*/
