#ifndef TIMEABS_H_
#define TIMEABS_H_

#include <common/Common.h>


/**
 * This time class is based on a non-monotonic clock.
 */


struct TimeAbs;
typedef struct TimeAbs TimeAbs;

static inline void TimeAbs_init(TimeAbs* this);
static inline void TimeAbs_initFromTimeval(TimeAbs* this, struct timeval* t);
static inline TimeAbs* TimeAbs_construct(void);
static inline TimeAbs* TimeAbs_constructFromTimeval(struct timeval* t);
static inline void TimeAbs_uninit(TimeAbs* this);
static inline void TimeAbs_destruct(TimeAbs* this);

static inline void TimeAbs_setToNow(TimeAbs* this);
static inline int TimeAbs_elapsedSinceMS(TimeAbs* this, TimeAbs* earlierT);
static inline int TimeAbs_elapsedSinceMicro(TimeAbs* this, TimeAbs* earlierT);
static inline int TimeAbs_elapsedMS(TimeAbs* this);

// getters & setters
static inline struct timeval* TimeAbs_getTimeval(TimeAbs* this);

// operators
static inline fhgfs_bool TimeAbs_equals(TimeAbs* this, TimeAbs* other);
static inline fhgfs_bool TimeAbs_notequals(TimeAbs* this, TimeAbs* other);


struct TimeAbs
{
   struct timeval now;
};


void TimeAbs_init(TimeAbs* this)
{
   do_gettimeofday(&this->now);
}

void TimeAbs_initFromTimeval(TimeAbs* this, struct timeval* t)
{
   this->now = *t;
}

TimeAbs* TimeAbs_construct()
{
   TimeAbs* this = (TimeAbs*)kmalloc(sizeof(*this), GFP_KERNEL);
   
   TimeAbs_init(this);
   
   return this;
}

TimeAbs* TimeAbs_constructFromTimeval(struct timeval* t)
{
   TimeAbs* this = (TimeAbs*)kmalloc(sizeof(*this), GFP_KERNEL);
   
   TimeAbs_initFromTimeval(this, t);
   
   return this;
}

void TimeAbs_uninit(TimeAbs* this)
{
}

void TimeAbs_destruct(TimeAbs* this)
{
   TimeAbs_uninit(this);
   
   kfree(this);
}

void TimeAbs_setToNow(TimeAbs* this)
{
   do_gettimeofday(&this->now);
}

int TimeAbs_elapsedSinceMS(TimeAbs* this, TimeAbs* earlierT)
{
   int secs = (this->now.tv_sec - earlierT->now.tv_sec) * 1000;
   int micros = (this->now.tv_usec - earlierT->now.tv_usec) / 1000;
   
   return secs + micros;
}

int TimeAbs_elapsedSinceMicro(TimeAbs* this, TimeAbs* earlierT)
{
   int secs = (this->now.tv_sec - earlierT->now.tv_sec) * 1000000;
   int micros = (this->now.tv_usec - earlierT->now.tv_usec);
   
   return secs + micros;
}

int TimeAbs_elapsedMS(TimeAbs* this)
{
   int elapsedMS;
   
   TimeAbs currentT;
   TimeAbs_init(&currentT);
   
   elapsedMS = TimeAbs_elapsedSinceMS(&currentT, this);
   
   TimeAbs_uninit(&currentT);
   
   return elapsedMS;
}

struct timeval* TimeAbs_getTimeval(TimeAbs* this)
{
   return &this->now;
}

fhgfs_bool TimeAbs_equals(TimeAbs* this, TimeAbs* other)
{
   return (this->now.tv_usec == other->now.tv_usec) && (this->now.tv_sec == other->now.tv_sec);
}

fhgfs_bool TimeAbs_notequals(TimeAbs* this, TimeAbs* other)
{
   return (this->now.tv_usec != other->now.tv_usec) || (this->now.tv_sec != other->now.tv_sec);
}


#endif /*TIMEABS_H_*/
